import java.util.Iterator;
import java.util.ListIterator;

/**
 * First pass at factoring.
 */
public final class Factor1 {

    private static boolean seenString;

    public static void combineTags(Node n, int i)
    {
	if(i>0) {
	    if(!n.isString()) {
		
		Iterator iter = n.getSubtrees();
		
		while(iter.hasNext()) {
		    combineTags((Node) (iter.next()), i-1);
		}
		
		/* after all children are done, attempt to combine */
		if(n.size() == 1) {
		    Node child = n.getFirstChild();
		    if(!child.isString()) {
			n.setTag(Tag.compose(n.getTag(), child.getTag()));
			n.copyChildren(child);
		    }
		}
	    }
	}
	
    }
    public static void combineTags(Node n)
    {
	combineTags(n, Integer.MAX_VALUE);
    }

    public static Node getNonString(ListIterator iter)
    {
	Node nonString = null;
	while(nonString == null && iter.hasNext()) {
	    Node temp = (Node) (iter.next());
	    if(temp.isString()) {
		seenString = true;
	    } else {
		nonString = temp;
	    }
	}
	return nonString;
    }

    /*
    private static final int WINDOW_SIZE=30;
    private static final int NUM_SIZES=10;
    private static final int NUM_COLORS=8;
    private static int[] last = new int[NUM_SIZES];
    private static int[] crosses = new int[WINDOW_SIZE];
    */

    public static void factor(Node n)
    {
	if(!n.isString()) {

	    ListIterator iter = n.getSubtrees();

	    while(iter.hasNext()) {
		Node temp = (Node) iter.next();
		factor(temp);
		combineTags(temp, 1);
	    }

	    if(n.size() > 1) {
		
	    simplefactor:
		{
		    iter = n.getSubtrees();
		    
		    Node prev = getNonString(iter);
		    if(prev == null) break simplefactor;
		    
		    while(iter.hasNext()) {
			Node curr = (Node) (iter.next());
			if(curr.isString()) {
			    prev = getNonString(iter);
			    continue;
			}
			
			/* prev and curr are consecutive nonstrings */
			int currTag = curr.getTag();
			int prevTag = prev.getTag();
			int interTag = Tag.inter(currTag, prevTag);

			if(interTag != 0) {
			    //System.err.println("factor!");
			    /*
			      left: " + prevTag +
			      " right: " + currTag + " inter: " +
			      interTag + " newLeft: " +
			      Tag.diff(prevTag, interTag) +
			      " newRight: " +
			      Tag.diff(currTag, interTag));
			    */
			    
			    Node newPrev = new Node(Tag.diff(prevTag,
							     interTag));
			    newPrev.copyChildren(prev);
			    
			    curr.setTag(Tag.diff(currTag, interTag));
			    
			    prev.setTag(interTag);
			    prev.clearChildren();
			    prev.add(newPrev);
			    prev.add(curr);
			    
			    iter.remove();
			} else {
			    prev = curr;
			}
		    }
		}
	    }
	}
    }

    public static void factorWide(Node n, int precontext, int depth)
    {
	if(!n.isString()) {

	    int context = Tag.compose(precontext, n.getTag());

	    ListIterator iter = n.getSubtrees();

	    if(depth > 1) {
		while(iter.hasNext()) {
		    Node temp = (Node) iter.next();
		    factorWide(temp, context, depth-1);
		    combineTags(temp, 1);
		}
	    }

	    if(n.size() > 1) {
		iter = n.getSubtrees();
		int contextBuild = Tag.bitSetOnly(context) | Tag.pl;
		int contextCost = Tag.cost(contextBuild);
		
		//System.err.print("\ncontextcost: " + contextCost + ", ");
		    
	    PLfactor:
		while(iter.hasNext()) {
		    
		    Node prev;
		    int prevTag;
		    int prevBits;
		    int prevColor;
		    int prevSize;
		    
		    /* search for someone whose cost to build is more than
		       the context (using a PL) */
		    do {
			prev = getNonString(iter);
		        if(prev == null) break PLfactor;
			prevTag = prev.getTag();
			prevBits = Tag.bitSetOnly(prevTag);
			prevColor = Tag.getColor(prevTag);
			prevSize = Tag.getSize(prevTag);
			//System.err.print("prevcost " + Tag.cost(prevBits)
				//	 + ", ");
		    } while(Tag.cost(prevBits) <= contextCost
			    && prevColor == -1
			    && prevSize == -1);
		    
		    int prevNextIndex = iter.nextIndex();
		    prevColor = (prevColor==-1) ? -2 : prevColor;
		    prevSize = (prevSize==-1) ? -2 : prevSize;
		    
		    /* search for someone whose cost to build the intersection
		       of it and prev is more than the context (with a PL) */

		    seenString = false;
		    
		searchForNext:
		    while(true) {
			Node next = getNonString(iter);
			if(next == null) {
			    int i;
			    for(i=iter.nextIndex(); i>prevNextIndex; --i) {
				iter.previous();
			    }
			    continue PLfactor;
			}

			int interTag = Tag.empty;
			int contextTag = Tag.empty;

			int nextTag = next.getTag();
			int nextBits = Tag.bitSetOnly(nextTag);
			int interBits = Tag.inter(prevBits, nextBits);
		    
			//System.err.print("intercost " + Tag.cost(interBits)
			//					 + ", ");

			if(Tag.cost(interBits) > contextCost) {
			    
			    /* do the transformation! */
			    //System.err.println("plshort!");

			    interTag = interBits;
			    contextTag = contextBuild;
			}

			int nextColor = Tag.getColor(nextTag);
			if(seenString) {
			    prevColor = -2;
			    prevSize = -2;
			}
			if(nextColor == prevColor) {
			    //System.err.println("color!");
			    
			    interTag = Tag.setColor(interTag, nextColor);
			    // contextTag not needed
			}

			int nextSize = Tag.getSize(nextTag);
			if(nextSize == prevSize) {
			    //System.err.println("size!" + nextSize + " " +
				//	       prevSize);
			    
			    interTag = Tag.setSize(interTag, nextSize);
			    // contextTag not needed
			}


			if(interTag == Tag.empty) {
			    continue searchForNext;
			}

			/* needs iter, prevNextIndex interTag,
			   contextTag, prev, next */

			Node interNode = new Node(interTag);
			Node newNode = new Node(contextTag);

			int nextNextIndex = iter.nextIndex();
			    
			int i;
			for(i=nextNextIndex; i>prevNextIndex; --i) {
			    iter.previous();
			}
			interNode.add((Node) iter.previous());
			iter.remove();
			
			for( ; i<nextNextIndex-1; ++i) {
			    Node copy = (Node) iter.next();
			    iter.remove();
			    newNode.add(copy);
			}
			interNode.add(newNode);
			interNode.add((Node) iter.next());
			iter.remove();
			
			iter.add(interNode);
			iter.previous();
			
			prev.setTag(prevTag & ~interTag);
			next.setTag(nextTag & ~interTag);

			factorWide(newNode, context, 1);
			
			continue PLfactor;
		    }
		}
	    }
	}
    }

    public static void factorWide(Node n)
    {
	factorWide(n, 0, Integer.MAX_VALUE);
    }
}
