import java.io.*;

class MyOutStream extends FilterOutputStream {


    public MyOutStream(OutputStream out)
    {
	super(out);
    }

    public void print(char[] s) throws IOException
    {
	int sz = s.length;
	int i;
	for(i=0; i<sz; ++i) {
	    write(s[i]);
	}
    }
    public void print(String s) throws IOException
    {
	int sz = s.length();
	int i;
	for(i=0; i<sz; ++i) {
	    write(s.charAt(i));
	}
    }
}


public final class Main {
    public static final MyOutStream out = new MyOutStream(System.out);

    private static long start_time = 0L ;
    private static long end_time = 0L ;
    private static boolean allprint = false ;
    private static Node n = null ;
    private static boolean nophase = false ;

    private static void printDebugging(String phaseName) throws java.io.IOException {
	end_time = System.currentTimeMillis();
	if(! nophase){
	    System.err.println(phaseName + ":   \t" + (end_time-start_time) + " ms");
	}
	start_time = end_time;

	if(allprint) {
	    n.print();
	    
	    end_time = System.currentTimeMillis();
	    System.err.println("printing:\t" + (end_time-start_time) + " ms");
	    start_time = end_time;
	}
    }

    public static void main(String[] s) throws IOException
    {
	int index=0;
	boolean print = true;
	boolean error = false;

	start_time = System.currentTimeMillis();
	
	while(index < s.length) {

	    if(s[index].equals("-noprint")) {
		index++;
		print = false;
		continue;
	    }
	    if(s[index].equals("-nophase")) {
		index++;
		nophase = true;
		continue;
	    }
	    if(s[index].equals("-allprint")) {
		index++;
		allprint = true;
		continue;
	    }

	    try {
		String currFile = s[index];
		index++;
		InputStream in =
		    new BufferedInputStream(new FileInputStream(currFile));
		n = new Parser(in).superParse();
		break;
	    } catch (FileNotFoundException e) {
		System.err.println("file not found, trying next");
		error = true ;
	    } catch (IOException e) {
		System.err.println("parse error in file, trying next");
		error = true ;
	    }
	}

	if(n==null) {
	    try {
		if(error){
		    System.err.println("no file found on command line");
		    System.err.println("taking input from stdin");
		}
		n = new Parser(System.in).superParse();
	    } catch (IOException e) {
		System.err.println("parse error on stdin, bailing");
		System.exit(1);
	    }
	}


	printDebugging("parsing");

	Peephole.peep(n);
	printDebugging("peephole");

	n = Clean.clean(n);
	printDebugging("clean");

	n = Hoist.hoist(n);
	printDebugging("hoisting");

	Factor1.combineTags(n);
	printDebugging("combining");

	n = Clean.clean(n);
	printDebugging("clean");

	Factor1.factor(n);
	printDebugging("factoring");

	n = Clean.clean(n);
	printDebugging("clean");
	
	Factor1.factorWide(n);
	printDebugging("factorwide");

	n = Clean.clean(n);
	printDebugging("clean");

	n = Clean.uncombineTags(n);
	printDebugging("uncombining");

	n = Hoist.hoist(n);
	printDebugging("hoisting");

	Factor1.combineTags(n);
	printDebugging("combining");

	n = Clean.clean(n);
	printDebugging("clean");

	n = Hoist.hoist(n);
	printDebugging("hoisting");

	Factor1.combineTags(n);
	printDebugging("combining");

	n = Clean.clean(n);
	printDebugging("clean");

	if(print) {
	    n.print();
	    printDebugging("printing");
	}

	Main.out.print("\n<END>\n");
	Main.out.flush();

    }

}
