import java.io.*;
import java.util.*;

class Exception extends RuntimeException {}

final public class Node {
    private int tag;

    /*
      For non-string nodes, this is a Node.
      For string nodes, this is a StringBuffer during parsing,
      but then a String.
    */
    private Object child1OrString;


    private Node child2;
    private Node child3;
    private Stack moreNodes;
    
    public Node(int tag)
    {
	this.tag = tag;
    }

    /** only used by parser */
    public Node(StringBuffer s, boolean allSpace)
    {
	if(allSpace){
	    tag = Tag.spaceString;
	}else{
	    tag = Tag.string;
	}
	child1OrString = s;
    }

    public boolean isString()
    {
	return Tag.isString(tag);
    }

    public String getString()
    {
	return (String) child1OrString;
    }
    /** only used by parser */
    public StringBuffer getStringBuffer()
    {
	return (StringBuffer) child1OrString;
    }
    public void setString(String s)
    {
	child1OrString = s.intern();
    }

    public boolean isTree()
    {
	return !Tag.isString(tag);
    }

    public int getTag()
    {
	return tag;
    }

    public void setTag(int tag)
    {
	this.tag = tag;
    }

    public ListIterator getSubtrees() 
    {
	return getSubtrees(0);
    }

    public ListIterator getSubtreesEnd()
    {
	return getSubtrees(size());
    }

    private ListIterator getSubtrees(final int index)
    {
	return new ListIterator() {
	    private int pos=index; 

	    private static final int NO_REMOVE = 0;
	    private static final int LAST_NEXT = 1;
	    private static final int LAST_PREV = 2;
	    private int removeOk;

	    public boolean hasX(int index)
		{
		    if(index==-1) {
			return false;
		    } else if(index==0) {
			return !isString() && child1OrString != null;
		    } else if(index==1) {
			return child2 != null;
		    } else if(index==2) {
			return child3 != null;
		    } else {
			if(moreNodes == null) return false;
			return index-3 < moreNodes.size();
		    }
		}
	    public boolean hasNext()
		{
		    return hasX(pos);
		}
	    public boolean hasPrevious()
		{
		    return hasX(pos-1);
		}
	    public Object getX(int index)
		{
		    if(index==0) {
			if (isString()) {
			    throw new NoSuchElementException();
			}
			if (child1OrString != null) {
			    return child1OrString;
			}
		    } else if(index==1) {
			if (child2 != null) {
			    return child2;
			}
		    } else if(index==2) {
			if (child3 != null) {
			    return child3;
			}
		    } else {
			if(moreNodes == null) {
			    throw new NoSuchElementException();
			}
			if(index-3 < moreNodes.size()) {
			    Object o = moreNodes.elementAt(index-3);
			    return o;
			}
		    }
		    throw new NoSuchElementException();
		}
	    public Object next()
		{
		    //System.err.println("pos in next: " + this + " " + pos);
		    Object o = getX(pos);
		    pos++;
		    removeOk = LAST_NEXT;
		    return o;
		}
	    public Object previous()
		{
		    //System.err.println("pos in previous: " + this + " " + pos);
		    if(pos>0) {
			pos--;
			Object o = getX(pos);
			removeOk = LAST_PREV;
			return o;
		    } else {
			throw new NoSuchElementException();
		    }
		}
	    public void remove()
		{
		    //System.err.println("pos in remove: " + this + " " + pos);
		    if(removeOk == NO_REMOVE) {
			throw new IllegalStateException();
		    }
		    if(removeOk == LAST_NEXT) {
			//System.out.println("ZZZZZZZZZZZZZZZZZZZZZZZZZZZ" + pos);
			pos--;
		    }

		    removeOk = NO_REMOVE;

		    Node temp=null;
		    
		    if(pos==0) {
			temp = child2;
			child1OrString = temp;
			if(temp == null) { return; }
			//System.out.println("AAAAA" + pos);
		    }
		    
		    if(pos<=1) {
			temp = child3;
			child2 = temp;
			if(temp == null) { return; }
		    }
		    
		    if(pos<=2) {
			temp = moreNodes==null ? null :
			    (moreNodes.size()==0 ? null :
			     (Node) moreNodes.elementAt(0));
			child3 = temp;
			if(temp == null) { return; }
			moreNodes.remove(0);
			return;
		    }
		    
		    moreNodes.remove(pos-3);
		}
	    public int nextIndex()
		{
		    return pos;
		}
	    public int previousIndex()
		{
		    return pos-1;
		}
	    public void set(Object o)
		{
		    if(removeOk == NO_REMOVE) {
			throw new IllegalStateException();
		    }

		    int p = pos;
		    if(removeOk == LAST_NEXT) {
			p--;
		    }

		    if(p==0) {
			child1OrString = o;
		    } else if(p==1) {
			child2 = (Node) o;
		    } else if(p==2) {
			child3 = (Node) o;
		    } else {
			moreNodes.set(p-3, o);
		    }
		}
	    public void add(Object o)
		{
		    //System.err.println("pos in add: " + this + " " + pos);
		    removeOk = NO_REMOVE;
		    
		    if(pos >= 3) {
			if(moreNodes == null) {
			    moreNodes = new Stack();
			}
			moreNodes.add(pos-3, o);
		    } else {
			if(child3 != null) {
			    if(moreNodes != null) {
				moreNodes.add(0, child3);
			    }else{
				moreNodes = new Stack ();
				moreNodes.push(child3);
			    }
			}
			if(pos==2) {
			    child3 = (Node) o;
			} else {
			    child3 = child2;
			    if(pos==1) {
				child2 = (Node) o;
			    } else {
				child2 = (Node) child1OrString;
				child1OrString = o;
			    }
			}
		    }
		    
		    pos++;
		}
	};
    }

    /** If 'this' and 'n' are strings, adds the string of 'n' to the
	end of that of 'this'.  Return true iff this happens. */
    private boolean mergeStrings(Node n)
    {
	if(isString() && n.isString()) {
	    if(child1OrString instanceof StringBuffer) {
		((StringBuffer) child1OrString).append
		    ((StringBuffer) (n.child1OrString));
	    } else if(child1OrString instanceof String) {
		child1OrString =
		    (String) child1OrString + (String) n.child1OrString;
	    } else {
		return false;
	    }

	    if(Tag.isSpaceString(tag) && !Tag.isSpaceString(n.tag)) {
		setTag(Tag.string);
	    }

	    return true;
	}
	return false;
    }

    public void add(Node n)
    {
	if(child1OrString == null) {
	    child1OrString = n;
	} else if(child2 == null) {
	    if(!((Node) child1OrString).mergeStrings(n)) {
		child2 = n;
	    }
	} else if(child3 == null) {
	    if(!(child2.mergeStrings(n))) {
		child3 = n;
	    }
	} else {
	    if(moreNodes == null) {
		moreNodes = new Stack();
	    }
	    moreNodes.push(n);
	}
    }

    public int size()
    {
	if(Tag.isString(tag)) {
	    return -1;
	}

	int sz=0;

	if(child1OrString == null) return sz;
	sz++;
	if(child2 == null) return sz;
	sz++;
	if(child3 == null) return sz;
	sz++;

	if(moreNodes == null) return sz;
	return sz + moreNodes.size();
    }

    /** does not check if it's a string */
    public Node getFirstChild()
    {
	return (Node) child1OrString;
    }
    public void copyChildren(Node n)
    {
	child1OrString = n.child1OrString;
	child2 = n.child2;
	child3 = n.child3;
	moreNodes = n.moreNodes;
    }
    public void clearChildren()
    {
	child1OrString = null;
	child2 = null;
	child3 = null;
	moreNodes = null;
    }
    public void clear()
    {
	setTag(Tag.empty);
	clearChildren();
    }

    private void printInternal() throws	IOException
    {
	if(isString()){
	    Main.out.print(getString());
	    return;
	}
	
	Tag.openToString(tag);

	Iterator I = getSubtrees();

	while(I.hasNext()){
	    ((Node) I.next()).printInternal();
	}

	Tag.closeToString(tag);
    }

    public void print() throws	IOException
    {
	printInternal();
	Main.out.print("\n<>\n");
	Main.out.flush();
    }
}
