import java.io.IOException;

final public class Tag {
    private static final int threeOneBits = (1 << 0) + (1 << 1) + (1 << 2) ;
    private static final int fourOneBits  = (1 << 0) + (1 << 1) + (1 << 2) + (1 << 3) ;

    /*
     * Tag definitions
     *
     */

    public static final int empty = 0 ;

    // simple formatting tags
    private static final int formatShift = 0 ;
    public static final int numFormatTags = 6 ;
    public static final int b  = 1 << (formatShift + 0) ;
    public static final int em = 1 << (formatShift + 1) ;
    public static final int i  = 1 << (formatShift + 2) ;
    public static final int pl = 1 << (formatShift + 3) ;
    public static final int s  = 1 << (formatShift + 4) ;
    public static final int tt = 1 << (formatShift + 5) ;

    public static final int plKill = (~0) ^ b ^ em ^ i ^ s ^ tt;

    // underline tags
    private static final int underlineShift = formatShift + numFormatTags ;
    public static final int numUnderlineLevels = 3 ;
    public static final int u1 = 1 << (underlineShift + 0) ;
    public static final int u2 = (1 << (underlineShift + 1)) ; //| u1 ;
    public static final int u3 = (1 << (underlineShift + 2)) ; //| u2 ;

    // size tags
    private static final int sizeShift = underlineShift + numUnderlineLevels ;
    private static final int sizeMask = fourOneBits << sizeShift ;
    private static final int numSizeBits = 4 ;
    public static final int s0 = 1 << sizeShift ;
    public static final int s1 = 2 << sizeShift ;
    public static final int s2 = 3 << sizeShift ;
    public static final int s3 = 4 << sizeShift ;
    public static final int s4 = 5 << sizeShift ;
    public static final int s5 = 6 << sizeShift ;
    public static final int s6 = 7 << sizeShift ;
    public static final int s7 = 8 << sizeShift ;
    public static final int s8 = 9 << sizeShift ;
    public static final int s9 = 10 << sizeShift ;
    private static final int dummySize = 11 << sizeShift ;

    // color tags
    private static final int colorShift = sizeShift + numSizeBits ;
    private static final int colorMask = fourOneBits << colorShift ;
    private static final int numColorBits = 4 ;
    public static final int rc = 1 << colorShift ;
    public static final int gc = 2 << colorShift ;
    public static final int bc = 3 << colorShift ;
    public static final int cc = 4 << colorShift ;
    public static final int mc = 5 << colorShift ;
    public static final int yc = 6 << colorShift ;
    public static final int kc = 7 << colorShift ;
    public static final int wc = 8 << colorShift ;
    private static final int dummyColor = 9 << colorShift ;

    private static final int miscShift = colorShift + numColorBits ;
    private static final int numMiscBits = 2 ;
    public static final int string = 1 << miscShift ;
    public static final int spaceString = 1 << (miscShift + 1) ;

    // reserved
    private static final int lexerShift = miscShift + numMiscBits;
    private static final int numLexerBits = 1 ;
    public static final int lexerExit = 1 << lexerShift ;
    public static final int lexerEOF  = 1 << lexerShift + 1 ;


    public static final int topLevel = dummySize | dummyColor ;
    public static final int noSpaceContext = ~1 ;

    /*
     * Tag predicates
     *
     */

    public static final boolean isString (int tag){
	return (tag == string) || (tag == spaceString) ;
    }

    public static final boolean isSpaceString (int tag){
	return tag == spaceString ;
    }

    public static final boolean isEmpty (int tag){
	return tag == empty ;
    }

    public static final boolean isB (int tag){
	return (tag & b) != 0 ;
    }

    public static final boolean isEm (int tag){
	return (tag & em) != 0 ;
    }

    public static final boolean isI (int tag){
	return (tag & i) != 0 ;
    }

    public static final boolean isPl (int tag){
	return (tag & pl) != 0 ;
    }

    public static final boolean isS (int tag){
	return (tag & s) != 0 ;
    }

    public static final boolean isTt (int tag){
	return (tag & tt) != 0 ;
    }

    public static final boolean isU1 (int tag){
	return (tag & u1) != 0 ;
    }

    public static final boolean isU2 (int tag){
	return (tag & u2) != 0 ;
    }

    public static final boolean isU3 (int tag){
	return (tag & u3) != 0 ;
    }

    // return the size field, if it has one.  otherwise, return -1
    public static final int getSize (int tag){
	return ((tag & sizeMask) >> sizeShift) - 1 ;
    }

    // return the color field, if it has one.  otherwise, return -1
    public static final int getColor (int tag){
	return ((tag & colorMask) >> colorShift) - 1 ;
    }

    /* 
     * Modification operations
     *
     */
    public static final int setSize (int t, int size){
	return (t & (~sizeMask)) | ((size + 1) << sizeShift) ;
    }

    public static final int setColor (int t, int color){
	return (t & (~colorMask)) | ((color + 1) << colorShift) ;
    }

    // turn a context into a space context
    public static final int spaceContext (int c){
	int c2 = (c & ~(s | em | i | b));

	if((c & (u1 | u2 | u3)) == 0){
	    c2 = setColor(c2, 7); // white
	}

	return c2 ;

    }


    /*
     * Operations across a pair of tags
     *
     */

    public static final int bitSetOnly (int t){
	return (~colorMask & ~sizeMask) & t ;
    }

    // intersection of characteristics of the two tags
    public static final int inter (int t1, int t2){
	int t = bitSetOnly(t1 & t2) ;

	// only one side has Pl?
	if(!isPl(t) && (isPl(t1) || isPl(t2))) {
	    t &= plKill;
	}

	// if only one side has S, then it won't have EM either,
	// so we're good

	// same color?
	int c;
	if((c = (t1 & colorMask)) == (t2 & colorMask)) {
	    t |= c;
	}

	// same size?
	int s;
	if((s = (t1 & sizeMask)) == (t2 & sizeMask)) {
	    t |= s;
	}

	return t ;

    }

    // set difference on the two tags
    // t2 MUST be the return from a call to inter that t1 was involved in
    public static final int diff (int t1, int t2){
	int t = t1 & ~t2;
	/*
        int t = bitSetOnly(t1 & ~t2);

	int c = getColor(t1) ;
	if( c != getColor(t2)){
	    t = setColor(t, c);
	}

	int s = getSize(t1) ;
	if( s != getSize(t2)){
	    t = setSize(t, s);
	}
	*/
	return t ;
    }

    // assuming t1 is outside of t2, combine two tags into one
    public static final int compose (int t, int t2){

	if((pl & t2) != 0){
	    // turn off all of these various flags, turn on pl
	    t = (t & ~(s | em | i | b | tt | u1 | u2 | u3)) | pl ;
	}

	if((s & t2) != 0){
	    // turn off em, turn on s
	    t = (t & ~em) | s ;
	}

	// add in b, i, tt, u1, u2, and u3 if on in t2
	t = t | ((b | i | tt | u1 | u2 | u3) & t2);

	// flip the EM state if t2 has EM on
	if(isEm(t2) && ! isS(t)){
	    t = t ^ em;
	}
	
	// set to the size in t2, if there is one
	int s = getSize(t2) ;
	if(s != -1){
	    t = setSize(t, s);
	}

	// set to the color in t2, if there is one
	int c = getColor(t2) ;
	if(c != -1){
	    t = setColor(t, c);
	}

	return t ;

    }


    // pl is not part of the context, so ignore it when comparing
    public static final int composeOntoContext (int t1, int t2){
	return compose(t1, t2) & ~pl ;
    }

    /*
     * String functions
     *
     */

    // return the length in characters of printing this tag
    public static final int cost
	(
	 int t
	 )
    {
	// based off of the print function
	int l = 0 ;

	if (isPl(t)){
	    l += 9 ;
	}

	if (isB(t)){
	    l += 7 ;
	}
	
	if (isEm(t)){
	    l += 9 ;
	}
	
	if (isI(t)){
	    l += 7 ;
	}
	    
	if (isS(t)){
	    l += 7 ;
	}
	    
	if (isTt(t)){
	    l += 9 ;
	}

	if (isU1(t)){
	    l += 7 ;
	}

	if (isU2(t)){
	    l += 7 ;
	}

	if (isU3(t)){
	    l += 7 ;
	}

	if(getSize(t) != -1){
	    l += 7 ;
	}

	if(getColor(t) != -1){
	    l += 7 ;
	}

	return l ;

    }

    private static final char openTag = '<';
    private static final char openClosingTag = '/';
    private static final char closeTag = '>';
    private static final char noSecond = (char) 0;

    private static final void singleTagString
	(char char1, char char2, boolean open) throws IOException {
	Main.out.write(openTag);
	if(!open)
	    Main.out.write(openClosingTag);
	Main.out.write(char1);
	if(char2 != 0)
	    Main.out.write(char2);
	Main.out.write(closeTag);
    }

    public static final void openToString (int t) throws IOException {
	if(!isEmpty(t)) {
	    // plain must be printed first
	    if (isPl(t)){
		singleTagString('P', 'L', true);
	    }

	    if (isB(t)){
		singleTagString('B', noSecond, true);
	    }
	    
	    if (isEm(t)){
		singleTagString('E', 'M', true);
	    }
	    
	    if (isI(t)){
		singleTagString('I', noSecond, true);
	    }
	    
	    if (isS(t)){
		singleTagString('S', noSecond, true);
	    }
	    
	    if (isTt(t)){
		singleTagString('T', 'T', true);
	    }

	    if (isU1(t)){
		singleTagString('U', noSecond, true);
	    }

	    if (isU2(t)){
		singleTagString('U', noSecond, true);
	    }

	    if (isU3(t)){
		singleTagString('U', noSecond, true);
	    }

	    int size = getSize(t);

	    if(size != -1){
		singleTagString((char) (size + '0'), noSecond, true);
	    }

	    switch (getColor(t)) {
	    case -1 : break ; // no color
	    case 0 : singleTagString('r', noSecond, true); break;
	    case 1 : singleTagString('g', noSecond, true); break;
	    case 2 : singleTagString('b', noSecond, true); break;
	    case 3 : singleTagString('c', noSecond, true); break;
	    case 4 : singleTagString('m', noSecond, true); break;
	    case 5 : singleTagString('y', noSecond, true); break;
	    case 6 : singleTagString('k', noSecond, true); break;
	    case 7 : singleTagString('w', noSecond, true); break;
	    }
	}else{
	    //singleTagString((char) '?', noSecond, true);
	}
    }

    public static final void closeToString (int t) throws IOException {
	if(!isEmpty(t)) {
	    switch (getColor(t)) {
	    case -1 : break ; // no color
	    case 0 : singleTagString('r', noSecond, false); break;
	    case 1 : singleTagString('g', noSecond, false); break;
	    case 2 : singleTagString('b', noSecond, false); break;
	    case 3 : singleTagString('c', noSecond, false); break;
	    case 4 : singleTagString('m', noSecond, false); break;
	    case 5 : singleTagString('y', noSecond, false); break;
	    case 6 : singleTagString('k', noSecond, false); break;
	    case 7 : singleTagString('w', noSecond, false); break;
	    }

	    int size = getSize(t);

	    if(size != -1){
		singleTagString((char) (size + '0'), noSecond, false);
	    }

	    if (isU1(t)){
		singleTagString('U', noSecond, false);
	    }

	    if (isU2(t)){
		singleTagString('U', noSecond, false);
	    }

	    if (isU3(t)){
		singleTagString('U', noSecond, false);
	    }

	    if (isTt(t)){
		singleTagString('T', 'T', false);
	    }

	    if (isS(t)){
		singleTagString('S', noSecond, false);
	    }
	    
	    if (isI(t)){
		singleTagString('I', noSecond, false);
	    }
	    
	    if (isEm(t)){
		singleTagString('E', 'M', false);
	    }
	    
	    if (isB(t)){
		singleTagString('B', noSecond, false);
	    }
	    
	    if (isPl(t)){
		singleTagString('P', 'L', false);
	    }
	}else{
	    //singleTagString((char) '?', noSecond, false);
	}
    }
}
